;===========================================
; Crystal Core v1.5
; By Erik557 (Requested by Daizo Dee Von)
;===========================================
; Description: A simple stationary boss that
; is attacked with the Marine Pop sprite. It
; shoots  aimed  projectiles  to the  player
; every few seconds. Health is configurable,
; among with some other options ;)
;===========================================

!Health   = 2          ; health in decimal, MINUS ONE.
!PrSprite = $03        ; projectile the sprite shoots.
!PjSpeed  = $2A        ; projectile's speed. max is 7f.
!MarPopN  = $00        ; marine pop's number.
!MPopProj = $01        ; marine pop's projectile number.
!InvTime  = $4F        ; invulnerability time.

!ShootPrj = 1          ; if the sprite should shoot projectiles or not.
                       ; 0 - no
                       ; 1 - yes
                       ; frequency configured below.

!Teleport = 1          ; teleport around the screen. 0 is not

!EndLevel = 0          ; if the level should end or not.
                       ; 0 - nothing OR teleport to a level (if you inserted the uberASM)
                       ; 1 - end level

!FixedPlc = 0          ; 0 - scrolls as normal
                       ; 1 - fixed within the screen
                       ; (note: 1 FORCES the sprite to appear on screen 0!)

!LockScrn = 1          ; 0 - doesn't lock the screen
                       ; 1 - locks the screen when the sprite appears
                       ; has unpredictable (?) behavior the above is 1!
                       
!ScrDis   = $00D1      ; how close to the edge to the screen the sprite should be
                       ; use only if LockScrn is 1! you don't want more than $00EF.
                       
!Unlock   = 1          ; has no effect if the above is 0
                       ; 0 - lets the screen locked
                       ; 1 - allows free camera movement when the sprite is defeated

Tilemap:
       db $80,$82,$A0,$A2   ; phase 1
       db $C4,$C6,$E4,$E6   ; phase 2
       db $C8,$CA,$E8,$EA   ; phase 3

if !ShootPrj == 1
ShootFrequency:                    ; only values 00,01,03,07,0f,1f,3f,7f,ff are allowed
       db $00,$1F,$1F,$1F,$1F      ; have as many entries as health!
       db $3F,$3F,$3F,$3F,$3F      ; 00 is, of course, no shooting.
       db $7F,$7F,$7F,$7F,$7F
endif

if !Teleport == 1
TeleportFreq:                      ; same as above
       db $00,$3F,$7F,$7F,$7F
       db $7F,$7F,$7F,$7F,$FF
       db $FF,$FF,$FF,$FF,$FF
endif

!HealthTbl    = $1528
!InvulnTbl    = $1540
!GFXIndex     = $1602
!XLo          = $E4
!XHi          = $14E0
!FixedXLo     = $1570
!YLo          = $D8
!YHi          = $14D4
!SprStatus    = $14C8


; Sprite INIT
print "INIT ",pc
       LDA.b #!Health
       STA !HealthTbl,x

if !FixedPlc == 1
       LDA !XLo,x
       STA !FixedXLo,x
endif

RTL

; Sprite MAIN
print "MAIN ",pc
       PHB
       PHK
       PLB
       JSR CrystalCore
       PLB
RTL

MainReturn:
       RTS

CrystalCore:
       LDA !HealthTbl,x
       CMP.b #!Health+1/3*2
       BCS .NoStore
       LDY #$04
       CMP.b #!Health+1/3
       BCS .Store4
       LDY #$08
.Store4
       TYA
       STA !GFXIndex,x

.NoStore
       JSR Graphics
       
       LDA $9D
       BNE MainReturn
       LDA !SprStatus,x
       CMP #$08
       BNE MainReturn

if !FixedPlc == 0
       LDY #$00
       JSR SubOffScreen
endif

if !FixedPlc == 1
       LDA $1A
       CLC
       ADC !FixedXLo,x
       STA !XLo,x
       LDA $1B
       ADC #$00
       STA !XHi,x
endif

if !LockScrn == 1
       LDA $143E
       BEQ .noScroll
       STA $1594,x

.noScroll
       LDA $1411
       BEQ .nh
       STA $1FD6,x

.nh
       LDA $1412
       BEQ .nv
       STA $151C,x

.nv    LDA $14E0,x
       XBA
       LDA $E4,x
       REP #$20
       SBC $1A
       CMP #!ScrDis
       SEP #$20
       BCS +

       STZ $143E
       STZ $1411
       STZ $1412
+
endif

       JSL $01A7DC
       BCC .NoContact
       JSL $00F5B7
.NoContact

       LDA !InvulnTbl,x
       BNE .Invulnerable
       JSR ProjInteract

.Invulnerable
       LDA !HealthTbl,x
       TAY

if !ShootPrj == 1
       LDA $14E0,x
       XBA
       LDA $E4,x
       REP #$20
       SBC $1A
       CMP #!ScrDis
       SEP #$20
       BCS .NoTele
       LDA ShootFrequency,y
       BEQ .NotYet
       LDA $14
       AND ShootFrequency,y
       BNE .NotYet
       if !Teleport == 1
              PHY
       endif
       JSR ShootProjectile
       if !Teleport == 1
              PLY
       endif
.NotYet
endif
if !Teleport == 1
       LDA TeleportFreq,y
       BEQ .NoTele
       LDA $14
       AND TeleportFreq,y
       BNE .NoTele
-      JSL $01ACF9
       LDA $148D
       AND #$03
       TAY
       LDA ypos,y
       CMP !YLo,x
       BEQ -
       STA !YLo,x
       LDA #$10
       STA $1DF9
endif

.NoTele
       RTS

if !Teleport == 1
ypos:  db $10,$30,$50,$70
endif

;========================
; Sprite Graphics Routine
;========================

offsets:
       db $01,$02,$03,$04,$04,$03,$02,$01
       db $FF,$FE,$FD,$FC,$FC,$FD,$FE,$FF
palette:
       db $05,$07,$09,$0B
Xdisp:
       db $00,$10,$00,$10
Ydisp:
       db $00,$00,$10,$10

Graphics:
       JSR GetDrawInfo
       
       LDA !GFXIndex,x
       STA $03

       PHX
       LDX #$03
.Loop
       LDA $00
       CLC
       ADC Xdisp,x
       STA $0300,y
       
       LDA $14
       LSR
       AND #$0F
       PHX
       TAX
       LDA $01
       CLC
       ADC offsets,x
       PLX
       CLC
       ADC Ydisp,x
       STA $0301,y
       
       PHX
       TXA
       CLC
       ADC $03
       TAX
       
       LDA Tilemap,x
       STA $0302,y
       PLX
       
       PHX
       LDX $15E9
       LDA !InvulnTbl,x
       BEQ .Flash
       LDA #$27
       STA $0303,y
       BRA +

.Flash       
       LDA $14
       LSR #5
       AND #$03
       PHX
       TAX
       LDA palette,x
       ORA $64
       STA $0303,y
       PLX
+      PLX
       
       INY #4
       DEX
       BPL .Loop
       
       PLX
       LDY #$02
       LDA #$03
       JSL $01B7B3
       RTS

;==================================
; Misc. Routines.
;==================================

ShootProjectile:
       PHX                  ; Push X register (current shooter slot number) into stack
       LDX #$0B             ; Load number of slots in X (-1 (11 or 21), because slot 0 is included)
.Loop
       LDA !SprStatus,x     ;\ If current slot occupied, next slot
       BEQ SetupSpr         ;/ Otherwise, if one empty is found, goto setup sprite (basic parts of it)
.NextSlot
       DEX                  ; Next slot index
       BPL .Loop            ; If X is 0 or positive, loop again until -1.
.ShootReturn
       PLX
       RTS
SetupSpr:
       LDA #!PrSprite       ;\ Set custom sprite number
       STA $7FAB9E,x        ;/
       INC !SprStatus,x     ; Initalize sprite
       JSL $07F7D2          ;\ Reset sprite tables (so if spawned, doesn't use leftover values used by another sprite when despawned)
       JSL $0187A7          ;/
       LDA #$08             ;\ Set extra bit
       STA $7FAB10,x        ;/
       TXY                  ; Transfer current sprite slot into Y.
       PLX                  ; Pull x to restore it to be the shooter's current slot
       LDA !XLo,x           ;\  set x position for new sprite
       STA.w !XLo,y         ; |
       LDA !XHi,x           ; |
       STA !XHi,y           ;/
       LDA !YLo,x           ;\  set y position for new sprite
       CLC                  ; | (y position of generator - 8)
       ADC #$08             ; |
       STA.w !YLo,y         ; |
       LDA !YHi,x           ; |
       ADC #$00             ; |
       STA !YHi,y           ;/

       LDA !XHi,x           ;\  Get horizontal distance between mario and the
       XBA                  ; | shooter for $00.
       LDA !XLo,x           ; |
       REP #$20             ; |
       SEC                  ; |
       SBC $94              ; |
       STA $00              ; |
       SEP #$20             ;/

       LDA !YHi,x          ;\  Same as above but vertical for $02
       XBA                  ; |
       LDA !YLo,x            ; |
       REP #$20             ; |
       SEC                  ; |
       SBC $96              ; |
       SEC                  ;  \  Don't aim for the upper half of mairo.
       SBC #$0010           ;  /
       STA $02              ; |
       SEP #$20             ;/
       LDA #!PjSpeed        ; Load speed into A
       JSR aiming           ; Start aiming
       LDA $00              ;\  Set speed.
       STA $00B6,y          ; |
       LDA $02              ; |
       STA $00AA,y          ;/
       
       LDA #$06
       STA $1DFC
	RTS

ProjInteract:
       LDX #$0B
-      LDA !SprStatus,x
       BEQ +
       LDA $7FAB9E,x
       CMP #!MPopProj
       BEQ .FreeLoop
       BRA +
.CheckNext
       TYX
+      DEX
       BPL -
       LDX $15E9
       RTS

.FreeLoop
       TXY
       LDX $15E9
       LDA $14E0,x
       XBA
       LDA $E4,x
       REP #$20
       SBC $1A
       CMP #!ScrDis
       SEP #$20
       BCS .Ret
       LDA.w !XLo,y
       SEC
       SBC #$07
       STA $00
       LDA !XHi,y
       SBC #$00
       STA $08
       LDA.b #17
       STA $02
       LDA.w !YLo,y
       STA $01
       LDA !YHi,y
       STA $09
       LDA.b #32
       STA $03
       JSL $03B69F
       JSL $03B72B
       BCC .CheckNext
       LDA #!InvTime
       STA !InvulnTbl,x
       LDA #$28
       STA $1DFC
       LDA #$00
       STA !SprStatus,y
       LDA !HealthTbl,x
       BEQ Death
       DEC !HealthTbl,x
       TYX
       JSR Smoke
       LDX $15E9
.Ret
       RTS

Death:
if !LockScrn == 1
       if !Unlock == 1
              LDA $14E0,x
              XBA
              LDA $E4,x
              REP #$20
              SBC $1A
              CMP #!ScrDis
              SEP #$20
              BCS +
              LDA $1FD6,x
              STA $1411
              LDA $151C,x
              STA $1412
              LDA $1594,x
              STA $143E
       +
       endif
endif
if !EndLevel == 1
       STZ !SprStatus,x
       LDA #$3C
       STA $1DFC
       JSL $07FC3B
       LDA #$80
       STA $1DFB
       LDA #$FF
       STA $1493
       LDA #$00
       STA $141C
       LDA #$0B
       STA $71
.Ret
       RTS
else
       LDA #$0D
       STA $9E,x
       JSL $07F7D2
       LDA #$40
       STA !InvulnTbl,x
       LDA #$01
       STA $1534,x
       LDA #$09
       STA $1DFC
       LDA $1686,x
       ORA #$08
       STA $1686,x
       SEC
       ROR $167A,x
       LDY #$0B
.Loop
       LDA.w $9E,y
       CMP $9E,x
       BEQ ++
       PHX
       TYX
       LDA $7FAB9E,x
       CMP #!MarPopN
       BEQ +
       LDA #$00
       STA $14C8,y
+      PLX
++     DEY
       BPL .Loop
.Ret
       RTS
endif

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;Aiming routine, by MarioE.
;Input:
;$00 = (shooter_x - target_x), 16-bit
;$02 = (shooter_y - target_y), 16-bit
;A = projectile speed
;Output:
;$00 = X speed
;$02 = Y speed
;NOTE: distance more than $0100 (256 in decimal) will glitch, unless you divide them by
;the same number (like both by 4) to go under 256 and it still works.
;
;MarioE's explanation: Suppose one fired the projectile with X speed dx, and Y speed dy.
;Then its speed would be sqrt(dx2+dy2). Thus, we can adjust its speed by multiplying by
;speed / sqrt(dx2+dy2). This routine calculates the reciprocal 1 / sqrt(dx2+dy2),
;multiplies by speed, then multiplies by either dx or dy.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
recip_sqrt_lookup:
       dw $0000,$FFFF,$B505,$93CD,$8000,$727D,$6883,$60C2
       dw $5A82,$5555,$50F4,$4D30,$49E7,$4700,$446B,$4219
       dw $4000,$3E17,$3C57,$3ABB,$393E,$37DD,$3694,$3561
       dw $3441,$3333,$3235,$3144,$3061,$2F8A,$2EBD,$2DFB
       dw $2D41,$2C90,$2BE7,$2B46,$2AAB,$2A16,$2987,$28FE
       dw $287A,$27FB,$2780,$270A,$2698,$262A,$25BF,$2557
       dw $24F3,$2492,$2434,$23D9,$2380,$232A,$22D6,$2285
       dw $2236,$21E8,$219D,$2154,$210D,$20C7,$2083,$2041
       dw $2000,$1FC1,$1F83,$1F46,$1F0B,$1ED2,$1E99,$1E62
       dw $1E2B,$1DF6,$1DC2,$1D8F,$1D5D,$1D2D,$1CFC,$1CCD
       dw $1C9F,$1C72,$1C45,$1C1A,$1BEF,$1BC4,$1B9B,$1B72
       dw $1B4A,$1B23,$1AFC,$1AD6,$1AB1,$1A8C,$1A68,$1A44
       dw $1A21,$19FE,$19DC,$19BB,$199A,$1979,$1959,$1939
       dw $191A,$18FC,$18DD,$18C0,$18A2,$1885,$1869,$184C
       dw $1831,$1815,$17FA,$17DF,$17C5,$17AB,$1791,$1778
       dw $175F,$1746,$172D,$1715,$16FD,$16E6,$16CE,$16B7
       dw $16A1,$168A,$1674,$165E,$1648,$1633,$161D,$1608
       dw $15F4,$15DF,$15CB,$15B7,$15A3,$158F,$157C,$1568
       dw $1555,$1542,$1530,$151D,$150B,$14F9,$14E7,$14D5
       dw $14C4,$14B2,$14A1,$1490,$147F,$146E,$145E,$144D
       dw $143D,$142D,$141D,$140D,$13FE,$13EE,$13DF,$13CF
       dw $13C0,$13B1,$13A2,$1394,$1385,$1377,$1368,$135A
       dw $134C,$133E,$1330,$1322,$1315,$1307,$12FA,$12ED
       dw $12DF,$12D2,$12C5,$12B8,$12AC,$129F,$1292,$1286
       dw $127A,$126D,$1261,$1255,$1249,$123D,$1231,$1226
       dw $121A,$120F,$1203,$11F8,$11EC,$11E1,$11D6,$11CB
       dw $11C0,$11B5,$11AA,$11A0,$1195,$118A,$1180,$1176
       dw $116B,$1161,$1157,$114D,$1142,$1138,$112E,$1125
       dw $111B,$1111,$1107,$10FE,$10F4,$10EB,$10E1,$10D8
       dw $10CF,$10C5,$10BC,$10B3,$10AA,$10A1,$1098,$108F
       dw $1086,$107E,$1075,$106C,$1064,$105B,$1052,$104A
       dw $1042,$1039,$1031,$1029,$1020,$1018,$1010,$1008
		
aiming:
       PHX
       PHY
       PHP
       SEP #$30
       STA $0F

       LDX #$00
       REP #$20
       LDA $00
       BPL .pos_dx
       EOR #$FFFF
       INC
       INX
       INX
       STA $00
.pos_dx
       SEP #$20
       STA $4202
       STA $4203
		
       NOP
       NOP
       NOP
       REP #$20
       LDA $4216
       STA $04
       LDA $02
       BPL .pos_dy
       EOR #$FFFF
       INC
       INX
       STA $02
.pos_dy
       SEP #$20
       STA $4202
       STA $4203
       STX $0E

       REP #$30
       LDA $04
       CLC
       ADC $4216
       LDY #$0000
       BCC .loop
       INY
       ROR
       LSR
.loop
       CMP #$0100
       BCC +
       INY
       LSR
       LSR
       BRA .loop
+      CLC
       ASL
       TAX
       LDA recip_sqrt_lookup,x

-      DEY
       BMI +
       LSR
       BRA -
+      SEP #$30

       STA $4202
       LDA $0F
       STA $4203
       NOP
       STZ $05
       STZ $07
       LDA $4217
       STA $04
       XBA
       STA $4202
       LDA $0F
       STA $4203

       REP #$20
       LDA $04
       CLC
       ADC $4216
       STA $04
       SEP #$20

       LDX #$02
-      LDA $04
       STA $4202
       LDA $00,x
       STA $4203

       NOP
       NOP
       NOP
       NOP

       LDA $4217
       STA $06
       LDA $05
       STA $4202
       LDA $00,x
       STA $4203

       REP #$20
       LDA $06
       CLC
       ADC $4216
       SEP #$20
		
       LSR $0E
       BCS +
       EOR #$FF
       INC
+      STA $00,x
       DEX
       DEX
       BPL -

       PLP
       PLY
       PLX
       RTS

       
Smoke:
       LDY #$03
-      LDA $17C0,y
       BEQ +
       DEY
       BPL -
       RTS

+      LDA #$01
       STA $17C0,y
       LDA #$1C     
       STA $17CC,y
       LDA !YLo,x
       STA $17C4,y
       LDA !XLo,x
       STA $17C8,y
       RTS
       
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; GetDrawInfo - Optimized version
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

GetDrawInfo:
	LDA !XHi,x
	XBA
	LDA !XLo,x
	REP #$20
	SEC
	SBC $1A
	STA $00
	CLC
	ADC.w #$0040
	CMP.w #$0180
	SEP #$20
	LDA $01
	BEQ +
	LDA #$01
+	STA $15A0,x
	TDC
	ROL A
	STA $15C4,x
	BNE .Invalid
	LDA !YHi,x
	XBA
	LDA $1662,x
	AND #$20
	BEQ .CheckOnce
.CheckTwice
	LDA !YLo,x
	REP #$21
	ADC.w #$001C
	SEC
	SBC $1C
	SEP #$20
	LDA !YHi,x
	XBA
	BEQ .CheckOnce
	LDA #$02
.CheckOnce
	STA $186C,x
	LDA !YLo,x
	REP #$21
	ADC.w #$000C
	SEC
	SBC $1C
	SEP #$21
	SBC #$0C
	STA $01
	XBA
	BEQ .OnScreenY
	INC $186C,x
.OnScreenY
	LDY $15EA,x
	RTS
		
.Invalid
	PLA
	PLA
	RTS

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; SubOffScreen - Optimized version
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Argument
; Y : type (#$00=X0, #$02=X1 ... #$0E=X7)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
	
SubOffScreen:
	LDA $15A0,x
	ORA $186C,x
	BEQ .Return2
	LDA $167A,x
	AND #$04
	BNE .Return2
	LDA $5B
	LSR A
	BCS .VerticalLevel
	LDA $D8,x
	ADC #$50
	LDA $14D4,x
	ADC #$00
	CMP #$02
	BCS .EraseSprite
	LDA $14E0,x
	XBA
	LDA $E4,x
	REP #$21
	ADC .AddTable,y
	SEC
	SBC $1A
	CMP .CmpTable,y
.Common
	SEP #$20
	BCC .Return2
.EraseSprite
	LDA $14C8,x
	CMP #$08
	BCC .KillSprite
	LDY $161A,x
	CPY #$FF
	BEQ .KillSprite
	LDA $1938,y
	AND #$FE
	STA $1938,y
.KillSprite
	STZ $14C8,x
.Return2
	RTS

.VerticalLevel
	LDA $13
	LSR A
	BCS .CheckY
	LDA $14E0,x
	XBA
	LDA $E4,x
	REP #$21
	ADC.w #$0040
	CMP.w #$0280
	BRA .Common
.CheckY
	LDA $14D4,x
	XBA
	LDA $D8,x
	REP #$20
	SBC $1C
	CLC
	ADC.w #$0070
	CMP.w #$01D0
	BRA .Common

.AddTable
	dw $0040,$0040,$0010,$0070
	dw $0090,$0050,$0080,$FFC0
.CmpTable
	dw $0170,$01E0,$01B0,$01D0
	dw $0230,$01B0,$0220,$0160  